import json
import yaml
import base64
import pulumi

from pulumi.output import Output
from pulumi_aws import iam, eks, ec2

from pydantic.dataclasses import dataclass, ConfigDict
from typing import Literal, Optional, List, Mapping

from theia import ResourceOptions
from theia.aws.containers import lt_data
from theia.kubernetes.provider import Provider
from theia.kubernetes.configmap import ConfigMap, MetadataArgs
from theia.aws import get_caller_identity, get_region
from theia.aws.security.kms import Key
from theia.aws.security.iam import Role, InlinePolicy, Policy
from theia.aws.containers.k8s.policies import (
    EKS_ROLE_ASSUME_ROLE_POLICY,
    DEFAULT_EKS_POLICY,
    EKS_POLICY_ARNS,
)
from theia.aws.containers.k8s.cluster import (
    K8sCluster,
    ClusterVpcArgs,
    GenerateKubeConfig,
)
from theia.aws.containers.k8s.nodegroup import (
    NodeGroup,
    NodeGroupSettings,
    NodeGroupTaint,
    ScalingConfig,
)
from theia.aws.containers.k8s.launch_template import (
    LaunchTemplate,
    LaunchTemplateSettings,
    TagSpecification,
    BlockDeviceArgs,
)
from theia.aws.monitoring.cloudwatch import LogGroup


@dataclass(config=ConfigDict(arbitrary_types_allowed=True))
class EksNodeGroup:
    name: str
    zones: List[str]
    spot_instance: bool
    instance_types: List[str]
    node_role_arn: str | Output[str]
    block_device_args: List[BlockDeviceArgs]
    scaling_config: ScalingConfig
    security_group_ids: List[str | Output[str]]
    update_ami: Optional[bool] = False
    tags: Optional[Mapping[str, str | Output[str]]] = None
    labels: Optional[Mapping[str, str | Output[str]]] = None
    taints: Optional[List[NodeGroupTaint]] = None
    bootstrap_commands: Optional[str] = None
    bottlerocket_ami: Optional[bool] = False

@dataclass(config=ConfigDict(arbitrary_types_allowed=True))
class EksSettings:
    override_eks_role: Optional[iam.Role] = None
    subnet_zones: Optional[List[str]] = None
    public_access_cidrs: Optional[List[str | Output[str]]] = None
    default_addons_to_remove: Optional[List[str]] = None
    nodegroups: List[EksNodeGroup] = None
    log_group_destroy: bool = False


@dataclass(config=ConfigDict(arbitrary_types_allowed=True))
class AuthConfigMapData:
    rolearn: str | Output[str]
    username: str
    groups: List[str]


class Eks(pulumi.ComponentResource):
    def __init__(
        self,
        resource_name: str,
        eks_version: str,
        subnet_ids: Mapping[str, str],
        deployment_type: Literal["public", "private"],
        create_kms_key: Optional[bool] = True,
        security_group_ids: Optional[List[str]] = None,
        settings: Optional[EksSettings] = EksSettings(),
        auth_data: Optional[List[AuthConfigMapData]] = None,
        tags: Optional[Mapping[str, str]] = None,
        opts: Optional[ResourceOptions] = None,
    ) -> None:
        super().__init__("theia:aws:eks", resource_name, None, opts)

        if not settings.override_eks_role:
            self.__eks_role(name=resource_name, tags=tags)
        else:
            self.eks_role = settings.override_eks_role

        self.__cluster_admin_role(name=resource_name, tags=tags)

        if create_kms_key:
            self.__kms_key(name=resource_name, tags=tags)

        self.__eks_log_group(
            resource_name,
            tags,
            create_kms_key,
            settings.log_group_destroy,
        )
        self.__eks_cluster(
            resource_name,
            tags,
            deployment_type,
            settings.subnet_zones,
            subnet_ids,
            create_kms_key,
            security_group_ids,
            settings,
            eks_version,
        )
        self.__aws_auth_config_map(resource_name, auth_data)

        self.nodegroups = []
        if settings.nodegroups:
            for nodegroup in settings.nodegroups:
                nodegroup = self.__create_nodegroup(
                    resource_name,
                    nodegroup,
                    eks_version,
                    subnet_ids,
                )
                self.nodegroups.append(nodegroup)

        GenerateKubeConfig(self.cluster)
        self.register_outputs({})

    def __eks_role(self, name: str, tags: Mapping[str, str]) -> None:
        self.eks_role = Role(
            f"{name}-eks-role",
            assume_role_policy=json.dumps(EKS_ROLE_ASSUME_ROLE_POLICY),
            inline_policies=[InlinePolicy(name=f"{name}-eks-policy", policy=json.dumps(DEFAULT_EKS_POLICY))],
            managed_policy_arns=EKS_POLICY_ARNS,
            tags=tags,
            opts=ResourceOptions(parent=self),
        )

    def __cluster_admin_role(self, name: str, tags: Mapping[str, str]) -> None:
        account_id = get_caller_identity().account_id
        region_name = get_region().name

        self.cluster_admin_role = Role(
            f"{name}-eks-cluster-admin-role",
            assume_role_policy=json.dumps(
                {
                    "Version": "2012-10-17",
                    "Statement": [
                        {
                            "Sid": "",
                            "Effect": "Allow",
                            "Principal": {
                                "AWS": f"arn:aws:iam::{account_id}:root",
                            },
                            "Action": "sts:AssumeRole",
                        }
                    ],
                }
            ),
            inline_policies=[
                InlinePolicy(
                    name=f"access-to-cluster-{name}-only",
                    policy=json.dumps(
                        {
                            "Version": "2012-10-17",
                            "Statement": [
                                {
                                    "Sid": f"AllowAccessOnlyTo{name.replace('-', '')}",
                                    "Effect": "Allow",
                                    "Action": "eks:*",
                                    "Resource": f"arn:aws:eks:{region_name}:{account_id}:cluster/{name}",
                                }
                            ],
                        }
                    ),
                )
            ],
            tags=tags,
            opts=ResourceOptions(parent=self),
        )

        assumed_role_policy = self.cluster_admin_role.arn.apply(
            lambda arn: json.dumps(
                {
                    "Version": "2012-10-17",
                    "Statement": [
                        {
                            "Action": ["sts:AssumeRole"],
                            "Effect": "Allow",
                            "Resource": [arn],
                        }
                    ],
                }
            )
        )

        _ = Policy(
            f"{name}-eks-assumed-role-policy",
            description=f"Policy to be assumed in order to gain cluster admin access to {name} EKS cluster",
            policy=assumed_role_policy,
            tags=tags,
            opts=ResourceOptions(parent=self),
        )

    def __kms_key(self, name: str, tags: Mapping[str, str]) -> None:
        self.kms_key = Key(
            resource_name=f"{name}-key",
            description="KMS Key for EKS Cluster Encryption",
            tags=tags,
        )

    def __eks_log_group(
        self,
        name: str,
        tags: Mapping[str, str],
        create_kms: bool,
        log_group_destroy: bool,
    ) -> None:
        self.log_group = LogGroup(
            "/aws/eks/" + name + "/cluster",
            None if not create_kms else self.kms_key.id,
            retention_in_days=7,
            tags=tags,
            skip_destroy=not log_group_destroy,
        )

    def __eks_cluster(
        self,
        name: str,
        tags: Mapping[str, str],
        deployment_type: Literal["public", "private"],
        subnet_zones: Optional[List[str]],
        subnet_ids_: Mapping[str, str],
        create_kms: bool,
        sg_ids: List[str],
        settings: EksSettings,
        eks_version: str,
    ) -> None:
        subnet_ids: List[str] = []
        if subnet_zones:
            for zone in subnet_zones:
                subnet_ids.append(subnet_ids_[zone])
        else:
            for _, val in subnet_ids_.items():
                subnet_ids.append(val)
        self.cluster = K8sCluster(
            f"{name}-cluster",
            role_arn=self.eks_role.arn,
            version=eks_version,
            subnet_ids=subnet_ids,
            cluster_vpc_args=ClusterVpcArgs(
                endpoint_public_access=(deployment_type == "public"),
                public_access_cidrs=settings.public_access_cidrs,
                endpoint_private_access=True,
                security_group_ids=sg_ids,
            ),
            enabled_cluster_log_types=[
                "api",
                "audit",
                "authenticator",
                "controllerManager",
                "scheduler",
            ],
            kms_key_arn=None if not create_kms else self.kms_key.arn,
            default_addons_to_remove=settings.default_addons_to_remove,
            tags=tags,
            opts=ResourceOptions(depends_on=[self.log_group]),
        )

        self.kubeconfig = GenerateKubeConfig(self.cluster)
        self.k8s_provider = Provider(f"{name}-provider", kubeconfig=self.kubeconfig)

    def __aws_auth_config_map(self, name: str, data: List[AuthConfigMapData] | None) -> None:
        auth_data = []
        admin_auth_data = Output.all(self.cluster_admin_role.arn).apply(
            lambda arn: {
                "rolearn": arn[0],
                "username": f"{name}:clusterAdmin",
                "groups": ["system:masters"],
            }
        )
        auth_data.append(admin_auth_data)
        if data:
            for d in data:
                auth_data.append(
                    {
                        "rolearn": d.rolearn,
                        "username": d.username,
                        "groups": d.groups,
                    }
                )

        auth_data_json = Output.all(auth_data).apply(lambda data: {"mapRoles": yaml.dump(data[0])})
        self.auth_config_map = ConfigMap(
            resource_name=f"{name}-aws-auth-cm",
            data=auth_data_json,
            immutable=True,
            metadata=MetadataArgs(name="aws-auth", namespace="kube-system"),
            opts=ResourceOptions(
                provider=self.k8s_provider,
                depends_on=[self.k8s_provider, self.cluster_admin_role],
            ),
        )

    @staticmethod
    def __fetch_eks_ami_name(eks_version: str, nodegroup: EksNodeGroup) -> str:
        if nodegroup.bottlerocket_ami:
            ami_name = f"bottlerocket-aws-k8s-{eks_version}-x86_64-*"
        else:
            ami_name = f"amazon-eks-node-{eks_version}-*"

        return ami_name

    @staticmethod
    def __fetch_eks_ami(ami_name: str) -> str:
        ami = ec2.get_ami(
            most_recent=True,
            owners=["amazon"],
            filters=[
                ec2.GetAmiFilterArgs(
                    name="name", values=[ami_name]
                )
            ],
        )
        return ami.id

    def __create_launch_template(
        self,
        name: str,
        nodegroup: EksNodeGroup,
        eks_version: str,
    ) -> ec2.LaunchTemplate:
        vpc_sg_ids = [self.cluster.vpc_config.cluster_security_group_id]
        vpc_sg_ids += nodegroup.security_group_ids
        tagspec = None
        if nodegroup.tags:
            tagspec = [
                TagSpecification(resource_type="instance", tag=nodegroup.tags),
                TagSpecification(resource_type="volume", tag=nodegroup.tags),
            ]

        ami_name = self.__fetch_eks_ami_name(eks_version = eks_version, nodegroup=nodegroup)

        if nodegroup.bottlerocket_ami:
            user_data = lt_data.BOTTLEROCKET_USER_DATA
        else:
          if nodegroup.bootstrap_commands:
              user_data = (
                  lt_data.SCRIPT_FORMAT
                  + nodegroup.bootstrap_commands
                  + lt_data.SCRIPT_BOUNDARY_END
                  + lt_data.BASE_USER_DATA
              )
          else:
              user_data = (
                  lt_data.SCRIPT_FORMAT
                  + lt_data.SCRIPT_BOUNDARY_END
                  + lt_data.BASE_USER_DATA
              )

        nodelabels = ""
        if nodegroup.labels:
            for key,value in nodegroup.labels.items():
                nodelabels += f'"{key}" = "{value}"'

        formatted_user_data = (
            Output.all(
                self.cluster.name,
                self.cluster.endpoint,
                self.cluster.certificate_authority.apply(lambda v: v.data),
                self.cluster.kubernetes_network_config.apply(lambda v: v.service_ipv4_cidr),
            )
            .apply(
                lambda args: user_data.format(
                    cluster_name_value=args[0],
                    api_server_url_value=args[1],
                    b64_cluster_ca_value=args[2],
                    service_ipv4_cidr_value=args[3],
                    node_labels=nodelabels
                )
            )
            .apply(lambda user_data: base64.b64encode(user_data.encode()).decode())
        )
        lt_args = LaunchTemplateSettings(
            block_device_mapping=nodegroup.block_device_args,
            user_data=formatted_user_data,
            vpc_security_group_ids=vpc_sg_ids,
            resource_tag_mapping=tagspec,
            image_id=self.__fetch_eks_ami(ami_name = ami_name),
        )
        return LaunchTemplate(
            resource_name=f"{nodegroup.name}-{name}",
            settings=lt_args,
            tags=nodegroup.tags,
            opts=ResourceOptions(
                ignore_changes=None if nodegroup.update_ami else ["imageId"],
                depends_on=[self.cluster],
            ),
        )

    def __create_nodegroup(
        self,
        name: str,
        nodegroup: EksNodeGroup,
        eks_version: str,
        subnets: Mapping[str, str],
    ) -> eks.NodeGroup:
        launch_template = self.__create_launch_template(name, nodegroup, eks_version)
        nodegroup_args = NodeGroupSettings(
            instance_types=nodegroup.instance_types,
            labels=nodegroup.labels,
            taints=nodegroup.taints,
            # ami_type="BOTTLEROCKET_x86_64" if nodegroup.bottlerocket_ami else None
        )

        subnet_ids: List[str] = []
        for zone in nodegroup.zones:
            subnet_ids.append(subnets[zone])

        return NodeGroup(
            resource_name=nodegroup.name,
            cluster_name=self.cluster.name,
            node_role_arn=nodegroup.node_role_arn,
            subnet_ids=subnet_ids,
            capacity_type="SPOT" if nodegroup.spot_instance else "ON_DEMAND",
            launch_template_version=launch_template.latest_version,
            launch_template_id=launch_template.id,
            scaling_config=nodegroup.scaling_config,
            settings=nodegroup_args,
            tags=nodegroup.tags,
            opts=ResourceOptions(
                ignore_changes=["scalingConfig"],
                depends_on=[launch_template, self.auth_config_map],
            ),
        )
