import pulumi

from pulumi_aws import ec2
from pulumi.output import Output

from dataclasses import field
from pydantic.dataclasses import dataclass, ConfigDict
from typing import List, Mapping, Optional, Literal

from theia import ResourceOptions
from theia.utils.logger import Logger
from theia.definitions import THEIA_VERSION
from theia._registry import StackRegistry, GlobalTags

logger = Logger("ec2")


@dataclass
class Metadata:
    http_endpoint: Optional[Literal["enabled", "disabled"]] = "enabled"
    http_tokens: Optional[Literal["optional", "required"]] = None
    http_put_response_hop_limit: Optional[int] = 3
    instance_metadata_tags: Optional[Literal["enabled", "disabled"]] = "disabled"


@dataclass
class BlockDeviceArgs:
    device_name: Optional[str] = None
    virtual_name: Optional[str] = None
    ebs_iops: Optional[int] = None
    volume_type: Optional[
        Literal["standard", "gp2", "gp3", "io1", "io2", "sc1", "st1"]
    ] = None
    volume_size: Optional[int] = None


@dataclass(config=ConfigDict(arbitrary_types_allowed=True))
class TagSpecification:
    resource_type: str
    tag: Mapping[str, str | Output[str]]


@dataclass(config=ConfigDict(arbitrary_types_allowed=True))
class LaunchTemplateSettings:
    block_device_mapping: Optional[List[BlockDeviceArgs]] = None
    cpu_credits: Optional[Literal["standard", "unlimited"]] = None
    ebs_optimized: Optional[bool] = None
    iam_instance_profile_arn: Optional[str | Output[str]] = None
    vpc_security_group_ids: Optional[List[str | Output[str]]] = None
    resource_tag_mapping: Optional[List[TagSpecification]] = None
    user_data: Optional[str | Output[str]] = None
    metadata: Optional[Metadata] = field(default_factory=lambda: Metadata())
    image_id: Optional[str | Output[str]] = None


class LaunchTemplate:
    def __new__(
        cls,
        resource_name: str,
        description: Optional[str] = None,
        default_version: Optional[int] = None,
        update_default_version: Optional[bool] = True,
        instance_type: Optional[str] = None,
        tags: Optional[Mapping[str, str]] = None,
        settings: Optional[LaunchTemplateSettings] = LaunchTemplateSettings(),
        opts: Optional[ResourceOptions] = None,
    ) -> ec2.LaunchTemplate:
        name = {"Name": resource_name}
        merged_tags = THEIA_VERSION | name | GlobalTags.get()
        if tags:
            merged_tags = tags | merged_tags

        block_device_args = None
        if settings.block_device_mapping:
            block_device_args = []
            for block in settings.block_device_mapping:
                block_device_args.append(
                    ec2.LaunchTemplateBlockDeviceMappingArgs(
                        device_name=block.device_name,
                        ebs=ec2.LaunchTemplateBlockDeviceMappingEbsArgs(
                            iops=block.ebs_iops,
                            volume_size=block.volume_size,
                            volume_type=block.volume_type,
                        ),
                        virtual_name=block.virtual_name,
                    )
                )

        tag_spec = None
        if settings.resource_tag_mapping:
            tag_spec = []
            for tag in settings.resource_tag_mapping:
                tag_spec.append(
                    ec2.LaunchTemplateTagSpecificationArgs(
                        resource_type=tag.resource_type, tags=tag.tag
                    )
                )

        launch_template_args = ec2.LaunchTemplateArgs(
            block_device_mappings=block_device_args,
            credit_specification=ec2.LaunchTemplateCreditSpecificationArgs(
                cpu_credits=settings.cpu_credits
            ),
            default_version=default_version,
            update_default_version=update_default_version,
            description=description,
            ebs_optimized=settings.ebs_optimized,
            instance_type=instance_type,
            tag_specifications=tag_spec,
            iam_instance_profile=ec2.LaunchTemplateIamInstanceProfileArgs(
                arn=settings.iam_instance_profile_arn
            ),
            vpc_security_group_ids=settings.vpc_security_group_ids,
            user_data=settings.user_data,
            metadata_options=ec2.LaunchTemplateMetadataOptionsArgs(
                http_endpoint=settings.metadata.http_endpoint,
                http_put_response_hop_limit=settings.metadata.http_put_response_hop_limit,
                http_tokens=settings.metadata.http_tokens,
                instance_metadata_tags=settings.metadata.instance_metadata_tags,
            ),
            image_id=settings.image_id,
            tags=merged_tags,
        )

        launch_template = ec2.LaunchTemplate(
            resource_name, name=resource_name, args=launch_template_args, opts=opts
        )

        pulumi.export(
            f"launch-template---{resource_name}",
            {
                "arn": launch_template.arn,
                "id": launch_template.id,
                "latest_version": launch_template.latest_version,
            },
        )

        return launch_template

    @staticmethod
    def import_(stack: str, resource_name: str) -> ec2.Vpc:
        stack_ref = StackRegistry.get(stack_name=stack)
        lt = stack_ref.get_output(name=f"launch-template---{resource_name}")
        return ec2.LaunchTemplate.get(resource_name=resource_name, id=lt["id"])
