import pulumi

from pulumi_aws import ec2
from pulumi.output import Output

from typing import Optional, Mapping, List
from pydantic.dataclasses import dataclass, ConfigDict

from theia import ResourceOptions
from theia.definitions import THEIA_VERSION
from theia._registry import StackRegistry, GlobalTags
from theia.utils.errors import TheiaConfigurationError


@dataclass(config=ConfigDict(arbitrary_types_allowed=True))
class Route:
    cidr_block: Optional[str | Output[str]] = None
    gateway_id: Optional[str | Output[str]] = None
    instance_id: Optional[str] = None
    nat_gateway_id: Optional[str | Output[str]] = None
    vpc_endpoint_id: Optional[str | Output[str]] = None
    vpc_peering_connection_id: Optional[str | Output[str]] = None


@dataclass(config=ConfigDict(arbitrary_types_allowed=True))
class RouteTableAssociations:
    name: str
    subnet_id: Optional[str | Output[str]] = None
    gateway_id: Optional[str | Output[str]] = None

    def __post_init__(self):
        if not self.gateway_id and not self.subnet_id:
            raise TheiaConfigurationError(
                "Either one of subnet_id or gateway_id should be provided"
            )
        if self.gateway_id and self.subnet_id:
            raise TheiaConfigurationError(
                "Only one of subnet_id or gateway_id should be provided"
            )


class RouteTable:
    def __new__(
        cls,
        resource_name: str,
        vpc_id: str | Output[str],
        routes: List[Route],
        associations: List[RouteTableAssociations] = None,
        tags: Optional[Mapping[str, str]] = None,
        opts: Optional[ResourceOptions] = None,
    ) -> ec2.RouteTable:
        name = {"Name": resource_name}
        merged_tags = THEIA_VERSION | GlobalTags.get() | name
        if tags:
            merged_tags = tags | merged_tags

        route_args = []
        for route in routes:
            route_args.append(
                ec2.RouteTableRouteArgs(
                    cidr_block=route.cidr_block,
                    gateway_id=route.gateway_id,
                    network_interface_id=route.instance_id,
                    nat_gateway_id=route.nat_gateway_id,
                    vpc_endpoint_id=route.vpc_endpoint_id,
                    vpc_peering_connection_id=route.vpc_peering_connection_id,
                )
            )

        route_table = ec2.RouteTable(
            resource_name, routes=route_args, vpc_id=vpc_id, tags=merged_tags, opts=opts
        )

        if associations:
            for association in associations:
                ec2.RouteTableAssociation(
                    f"{association.name}-association",
                    route_table_id=route_table.id,
                    subnet_id=association.subnet_id,
                    gateway_id=association.gateway_id,
                    opts=ResourceOptions(parent=route_table, depends_on=route_table),
                )

        pulumi.export(
            f"route_table---{resource_name}",
            {"id": route_table.id, "arn": route_table.arn},
        )

        return route_table

    @staticmethod
    def import_(stack: str, resource_name: str) -> ec2.RouteTable:
        stack_ref = StackRegistry.get(stack_name=stack)
        rt = stack_ref.get_output(name=f"route_table---{resource_name}")
        return ec2.RouteTable.get(resource_name=resource_name, id=rt["id"])
