import pulumi
import pulumi_kubernetes as k8s

from pydantic.dataclasses import dataclass
from typing import Optional, Literal

from theia import ResourceOptions
from theia._registry import StackRegistry


@dataclass
class HelmSettings:
    driver: Optional[Literal["configmap", "secret", "memory", "sql"]] = None
    plugins_path: Optional[str] = None
    registry_config_path: Optional[str] = None
    repository_cache: Optional[str] = None
    repository_config_path: Optional[str] = None


@dataclass
class K8sClientSettings:
    burst: Optional[int] = 10
    qps: Optional[int] = 5
    timeout: Optional[int] = 32


class Provider:
    def __new__(
        cls,
        resource_name: str,
        cluster: Optional[str] = None,
        context: Optional[str] = None,
        kubeconfig: Optional[str] = None,
        default_namespace: Optional[str] = None,
        delete_unreachable: Optional[bool] = False,
        helm_settings: Optional[HelmSettings] = HelmSettings(),
        k8s_client_settings: Optional[K8sClientSettings] = K8sClientSettings(),
        opts: Optional[ResourceOptions] = None,
    ) -> k8s.Provider:
        helm_args = k8s.HelmReleaseSettingsArgs(
            driver=helm_settings.driver,
            plugins_path=helm_settings.plugins_path,
            registry_config_path=helm_settings.registry_config_path,
            repository_cache=helm_settings.repository_cache,
            repository_config_path=helm_settings.repository_config_path,
        )

        client_args = k8s.KubeClientSettingsArgs(
            burst=k8s_client_settings.burst,
            qps=k8s_client_settings.qps,
            timeout=k8s_client_settings.timeout,
        )

        provider = k8s.Provider(
            resource_name,
            cluster=cluster,
            context=context,
            kubeconfig=kubeconfig,
            namespace=default_namespace,
            delete_unreachable=delete_unreachable,
            helm_release_settings=helm_args,
            kube_client_settings=client_args,
            opts=opts,
        )

        pulumi.export(f"kubernetes_provider---{resource_name}", {"id": provider.id})
        return provider

    @staticmethod
    def import_(stack: str, resource_name: str) -> k8s.Provider:
        stack_ref = StackRegistry.get(stack_name=stack)
        provider = stack_ref.get_output(name=f"kubernetes_provider---{resource_name}")
        return k8s.Provider.get(resource_name=resource_name, id=provider["id"])
